/*
    This file is part of the lin-kd-tree library.
    Copyright (C) 2009  Laurent Poirrier

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "quicksort.h"
#include "kd-tree.h"
#include "space.h"
#include "array.h"

#define KDTREE_ALLOC_OVERHEAD	1024

#define DEBUG(s...)		fprintf(stderr, s)

// ************************************************************************
// 
// ************************************************************************
static int kdtree_array_resize(kdtree_array *b, KDTREE_SIZE sz)
{
	void *np;
	
	np = realloc(b->tree, sz * sizeof(kdtree_object));
	if (!np) {
		DEBUG("kdtree_array: malloc() failed for %d nodes\n", sz);
		return(-1);
	}
	b->tree = (kdtree_object *)np;
	b->nobj_alloc = sz;

	return(0);
}


// ************************************************************************
// 
// ************************************************************************
kdtree_array *kdtree_array_new()
{
	kdtree_array *b;
	
	b = (kdtree_array *)malloc(sizeof(kdtree_array));
	if (!b) {
		DEBUG("kdtree_array_new(): malloc() failed!\n");
		return(NULL);
	}
	
	b->nobj_tree = b->nobj_alloc = 0;
	b->tree = NULL;
	
	return(b);
}

void kdtree_array_free(kdtree_array *b)
{
	if (b) {
		if (b->tree) free(b->tree);
		free(b);
	}
}


void kdtree_clear(kdtree_array *b, KDTREE_SIZE base_size)
{
	if (b->nobj_alloc < base_size * 8) {
		// notes:
		// 1. this may fail without risks (base_size is only a hint)
		// 2. we should free/malloc instead of realloc
		kdtree_array_resize(b, base_size * 8);
	}
	
	b->nobj_tree = 0;
}



// ************************************************************************
// 
// ************************************************************************
int  kdtree_add_object(kdtree_array *b, kdtree_iaabb *bb, void *obj)
{
	kdtree_node xk[8];
	int n, nx, j;
	
	if (bb) {
		nx = kdtree_split_place(xk, bb, 1, 1, 1);
	} else {
		// convention for out of bound objects -- place at root
		kdtree_root(xk);
		nx = 1;
	}
	
	n = b->nobj_tree;
	
	if (n + nx > b->nobj_alloc) {
		if (kdtree_array_resize(b, n + nx + KDTREE_ALLOC_OVERHEAD))
			return(-1);
	}
	
	//kdtree_iaabb_dump(bb);
	//DEBUG("nx = %d\n", nx);
	
	for (j = 0; j < nx; j++) {
		b->tree[n].obj = obj;
		memcpy(&b->tree[n].path, &xk[j], sizeof(kdtree_node));
		
		//kdtree_dump(&xk[j]);
		
		n++;
	}
	
	b->nobj_tree += nx;
	return(0);
}

// **********************************************************
int  kdtree_add_aabb(kdtree_array *b, kdtree_space *s,
	kdtree_aabb *bb, void *obj)
{
	kdtree_iaabb ibb;
	
	if (kdtree_pack_aabb(s, &ibb, bb)) {
		// DEBUG("AABB out of bounds. Placing at root.\n");
		return(kdtree_add_object(b, NULL, obj));
	}
	
	return(kdtree_add_object(b, &ibb, obj));
}

// **********************************************************
int  kdtree_add_sphere(kdtree_array *b, kdtree_space *s,
	kdtree_sphere *bs, void *obj)
{
	kdtree_iaabb ibb;

	if (kdtree_pack_sphere(s, &ibb, bs)) {
		// DEBUG("Sphere out of bounds. Placing at root.\n");
		return(kdtree_add_object(b, NULL, obj));
	}
	
	return(kdtree_add_object(b, &ibb, obj));
}


// ************************************************************************
// Time step simulation -- kd-tree version
// ************************************************************************
void kdtree_collisions(kdtree_array *b, kdtree_collision_callback cb)
{
	kdtree_object *ip, *jp;
	int i, j;
	// int exam = 0;
	
	if ((!b) || (!b->tree)) return;

	kdtree_quicksort(b->tree, 0, b->nobj_tree - 1);

	i = 0;
	ip = b->tree;
	while (i < b->nobj_tree) {
		j = i + 1;
		jp = ip + 1;
		
		while ((j < b->nobj_tree)
		&&  (KDTREE_INTERSECT(ip->path, jp->path))) {
			// exam++;
			
			cb(ip->obj, jp->obj);
			
			jp++;
			j++;
		}
		
		ip++;
		i++;
	}
	
	/*
	DEBUG("%d / max %d interactions examined\n",
		exam, (b->nobj_tree * (b->nobj_tree - 1)) / 2);
	// */
}



