/*
    This file is part of the lin-kd-tree library.
    Copyright (C) 2009  Laurent Poirrier

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef __KDTREE_NODES_H
#define __KDTREE_NODES_H
#include <stdint.h>

/*
  The kd-tree node functions handle node path determination given
  integer-mapped coordinates, and operations on node paths.
*/

/*
 - types must be unsigned
 - KDTREE_COORD_BITS * 3 >= KDTREE_PATH_ELEMS * KDTREE_PATH_ELEM_BITS
*/

// *********************************************************************
// Octree node path representation
// *********************************************************************
#define KDTREE_COORD		uint32_t
#define KDTREE_COORD_BITS	32

#define KDTREE_PATH_ELEM_TYPE	uint32_t
#define KDTREE_PATH_ELEM_BITS	32
#define KDTREE_PATH_ELEM_LOG2	5
#define KDTREE_PATH_ELEMS	1

typedef struct kdtree_node_t {
	KDTREE_PATH_ELEM_TYPE path[KDTREE_PATH_ELEMS];
	int depth;
} kdtree_node;

typedef struct kdtree_iaabb_t {
	KDTREE_COORD x0, x1, y0, y1, z0, z1;
} kdtree_iaabb;



// *********************************************************************
// Functions
// *********************************************************************
void kdtree_init();

int  kdtree_before(kdtree_node *a, kdtree_node *b);
int  kdtree_intersect(kdtree_node *a, kdtree_node *b);
int  kdtree_ordered_intersect(kdtree_node *a, kdtree_node *b);

void kdtree_root(kdtree_node *dst);
void kdtree_place(kdtree_node *dst, kdtree_iaabb *bb);

int  kdtree_split_place(kdtree_node *dst8, kdtree_iaabb *bb,
	int split_x, int split_y, int split_z);


// debug
void kdtree_node_iaabb(kdtree_node *k, kdtree_iaabb *dst);

void kdtree_print(const char *name, const kdtree_node *a);
#define kdtree_dump(a)	kdtree_print(#a, a)

void kdtree_iaabb_print(const char *name, const kdtree_iaabb *bb);
#define kdtree_iaabb_dump(bb)	kdtree_iaabb_print(#bb, bb)

void kdtree_bin_print(const char *name, KDTREE_COORD c);
#define kdtree_bin_dump(c)	kdtree_bin_print(#c, c)

// *********************************************************************
// Inline predicates and accessors
// *********************************************************************

#define KDTREE_COORD_MSB	((KDTREE_COORD)1 << (KDTREE_COORD_BITS - 1))
#define KDTREE_COORD_ONES	(~(KDTREE_COORD)0)

#define KDTREE_PATH_ELEM_MSB	\
	((KDTREE_PATH_ELEM_TYPE)1 << (KDTREE_PATH_ELEM_BITS - 1))
#define KDTREE_PATH_ELEM_ONES	\
	(~(KDTREE_PATH_ELEM_TYPE)0)

// These defines take actual nodes in argument, not pointers

#define KDTREE_MIN_DEPTH(a, b)	\
	( ((a).depth < (b).depth) ? (a).depth : (b).depth )

// accelerators
#define KDTREE_SIMPLE_PATH	(KDTREE_PATH_ELEMS == 1)

#if KDTREE_SIMPLE_PATH

#define KDTREE_BEFORE(a, b)	\
	(   ((a).path[0]  < (b).path[0]) \
	|| (((a).path[0] == (b).path[0]) && ((a).depth < (b).depth)) )

#define KDTREE_INTERSECT(a, b)	\
	(!( ((a).path[0] ^ (b).path[0]) >> \
		( KDTREE_PATH_ELEM_BITS - KDTREE_MIN_DEPTH(a, b) ) \
	) )

#define KDTREE_ORDERED_INTERSECT(a, b) \
	( ((a).depth <= (b).depth) && (!( ((a).path[0] ^ (b).path[0]) >> \
		(KDTREE_PATH_ELEM_BITS - (a).depth) )) )
#else
// function fallbacks

#define KDTREE_BEFORE(a, b)	\
	kdtree_before(&(a), &(b))

#define KDTREE_INTERSECT(a, b)	\
	kdtree_intersect(&(a), &(b))

#define KDTREE_ORDERED_INTERSECT(a, b) \
	kdtree_ordered_intersect(&(a), &(b))

#endif



#endif


