/*
    This file is part of the lin-kd-tree library.
    Copyright (C) 2009  Laurent Poirrier

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "vec_math.h"
#include "gen_graph.h"

#define PI	3.1415926535897932384626433832795029
#define SQRT2	1.4142135623730950488016887242096981
#define SQRT3	1.7320508075688772935274463415058724

// *******************************************************************
// graph alloc
// *******************************************************************
graph_surface *graph_alloc(int nv, int ne, int nf)
{
	graph_surface *r;
	void *ptr;

	ptr = malloc(sizeof(graph_surface)
		+ nv * sizeof(vec3)
		+ ne * sizeof(struct graph_edge_t)
		+ nf * sizeof(struct graph_face_t) );
	if (!ptr) return(NULL);

	r = ptr;
	ptr += sizeof(graph_surface);

	r->vertices = ptr;
	ptr += nv * sizeof(vec3);

	r->edges = ptr;
	ptr += ne * sizeof(struct graph_edge_t);

	r->faces = ptr;
	//ptr += nf * sizeof(struct graph_face_t);

	r->nvertices = nv;
	r->nedges = ne;
	r->nfaces = nf;
	return(r);
}


// *******************************************************************
// split every face in 4 subfaces / split every edge in 2 subedges
// *******************************************************************
graph_surface *graph_split(graph_surface *o)
{
	graph_surface *r;
	int e, f, vi, ei;
	int e0a, e1a, e2a, e0b, e1b, e2b;
	int e0mv, e1mv, e2mv;
	int v0, v1;

	r = graph_alloc(o->nvertices + o->nedges,
		o->nedges * 2 + o->nfaces * 3, o->nfaces * 4);
	if (!r) return(NULL);

	memcpy(r->vertices, o->vertices, o->nvertices * sizeof(vec3));
	vi = o->nvertices;

	// split edges
	for (e = 0; e < o->nedges; e++) {
		v0 = o->edges[e].v[0];
		v1 = o->edges[e].v[1];

		vec3_center(&r->vertices[vi],
			&r->vertices[v0], &r->vertices[v1]);

		r->edges[(e<<1)+0].v[0] = v0;
		r->edges[(e<<1)+0].v[1] = vi;
		r->edges[(e<<1)+1].v[0] = vi;
		r->edges[(e<<1)+1].v[1] = v1;

		vi++;
	}
	ei = o->nedges << 1;

	// split faces
	for (f = 0; f < o->nfaces; f++) {
		e0a = o->faces[f].e[0] << 1;
		e1a = o->faces[f].e[1] << 1;
		e2a = o->faces[f].e[2] << 1;

		e0mv = r->edges[e0a].v[1];
		e1mv = r->edges[e1a].v[1];
		e2mv = r->edges[e2a].v[1];
		r->edges[ei+0].v[1] = r->edges[ei+1].v[0] = e0mv;
		r->edges[ei+1].v[1] = r->edges[ei+2].v[0] = e1mv;
		r->edges[ei+0].v[0] = r->edges[ei+2].v[1] = e2mv;

		if (r->edges[e0a].v[0] == o->faces[f].v[0]) e0b = e0a + 1;
		else e0b = e0a++;
		if (r->edges[e1a].v[0] == o->faces[f].v[1]) e1b = e1a + 1;
		else e1b = e1a++;
		if (r->edges[e2a].v[0] == o->faces[f].v[2]) e2b = e2a + 1;
		else e2b = e2a++;

		r->faces[(f<<2)+0].e[0] = e0a;
		r->faces[(f<<2)+0].e[1] = ei+0;
		r->faces[(f<<2)+0].e[2] = e2b;
		r->faces[(f<<2)+1].e[0] = e0b;
		r->faces[(f<<2)+1].e[1] = e1a;
		r->faces[(f<<2)+1].e[2] = ei+1;
		r->faces[(f<<2)+2].e[0] = ei+2;
		r->faces[(f<<2)+2].e[1] = e1b;
		r->faces[(f<<2)+2].e[2] = e2a;
		r->faces[(f<<2)+3].e[0] = ei+0;
		r->faces[(f<<2)+3].e[1] = ei+1;
		r->faces[(f<<2)+3].e[2] = ei+2;

		r->faces[(f<<2)+0].v[0] = o->faces[f].v[0];
		r->faces[(f<<2)+0].v[1] = e0mv;
		r->faces[(f<<2)+0].v[2] = e2mv;
		r->faces[(f<<2)+1].v[0] = e0mv;
		r->faces[(f<<2)+1].v[1] = o->faces[f].v[1];
		r->faces[(f<<2)+1].v[2] = e1mv;
		r->faces[(f<<2)+2].v[0] = e2mv;
		r->faces[(f<<2)+2].v[1] = e1mv;
		r->faces[(f<<2)+2].v[2] = o->faces[f].v[2];
		r->faces[(f<<2)+3].v[0] = e2mv;
		r->faces[(f<<2)+3].v[1] = e0mv;
		r->faces[(f<<2)+3].v[2] = e1mv;

		ei += 3;
	}
	return(r);
}


// *******************************************************************
// 
// *******************************************************************

// Great care must be taken here for face/edge ordering
static vec3 tetra_vertices[4] = {
	{{ 2.0*SQRT2/3.0,  -1.0/3.0,  .0 }},
	{{ -SQRT2/3.0,     -1.0/3.0,  -SQRT2*SQRT3/3.0 }},
	{{ -SQRT2/3.0,     -1.0/3.0,   SQRT2*SQRT3/3.0 }},
	{{ .0,             1.0,       .0 }} };
static struct graph_edge_t tetra_edges[6] = {
	{{ 0, 1 }}, {{ 0, 2}}, {{ 0, 3}},
	{{ 1, 2 }}, {{ 1, 3}}, {{ 2, 3}} };
static struct graph_face_t tetra_faces[4] = {
	{ { 0, 2, 1 }, { 1, 3, 0} },
	{ { 0, 1, 3 }, { 0, 4, 2} },
	{ { 1, 2, 3 }, { 3, 5, 4} },
	{ { 2, 0, 3 }, { 1, 2, 5} } };

graph_surface *graph_gen_sphere(int divlevel)
{
	graph_surface *o;
	graph_surface *r;
	int t, v;

	r = graph_alloc(4, 6, 4);
	memcpy(r->vertices, tetra_vertices, sizeof(tetra_vertices));
	memcpy(r->edges, tetra_edges, sizeof(tetra_edges));
	memcpy(r->faces, tetra_faces, sizeof(tetra_faces));
	o = r;

	for (t = 0; t < divlevel; t++) {
		r = graph_split(o);
		free(o);
		if (!r) return(NULL);
		for (v = 0; v < r->nvertices; v++)
			vec3_normalize(&r->vertices[v]);
		o = r;
	}

	return(r);
}

// *******************************************************************
// 
// *******************************************************************

static vec3 sheet_vertices[4] = {
	{{ -1.0,  .0,  -1.0 }},
	{{ -1.0,  .0,   1.0 }},
	{{  1.0,  .0,   1.0 }},
	{{  1.0,  .0,  -1.0 }} };
static struct graph_edge_t sheet_edges[5] = {
	{{ 0, 1 }}, {{ 0, 2}}, {{ 0, 3}}, {{ 1, 2 }}, {{ 2, 3}} };
static struct graph_face_t sheet_faces[2] = {
	{ { 0, 1, 2 }, { 0, 3, 1} },
	{ { 0, 2, 3 }, { 1, 4, 2} } };

graph_surface *graph_gen_sheet(int divlevel)
{
	graph_surface *o;
	graph_surface *r;
	int t;

	r = graph_alloc(4, 5, 2);
	memcpy(r->vertices, sheet_vertices, sizeof(sheet_vertices));
	memcpy(r->edges, sheet_edges, sizeof(sheet_edges));
	memcpy(r->faces, sheet_faces, sizeof(sheet_faces));
	o = r;

	for (t = 0; t < divlevel; t++) {
		r = graph_split(o);
		free(o);
		if (!r) return(NULL);
		o = r;
	}

	return(r);
}

// *******************************************************************
// face normals
// *******************************************************************
vec3 *graph_face_normals(graph_surface *r)
{
	vec3 v1, v2, vc;
	vec3 *n, *v0;
	int t;
	
	n = malloc(r->nfaces * sizeof(vec3));
	if (!n) return(NULL);
	
	for (t = 0; t < r->nfaces; t++) {
		v0 = &r->vertices[r->faces[t].v[0]];
		
		vec3_difference(&v1, &r->vertices[r->faces[t].v[1]], v0);
		vec3_difference(&v2, &r->vertices[r->faces[t].v[2]], v0);
		
		vec3_cross_product(&vc, &v1, &v2);
		vec3_normalize(&vc);
		vec3_set(&n[t], &vc);
	}
		
	return(n);
}

// *******************************************************************
// mean vertex normals
// *******************************************************************
vec3 *graph_vertex_normals(graph_surface *r)
{
	vec3 v1, v2, vc;
	int v0n, v1n, v2n;
	vec3 *n;
	int t, i;
	
	n = malloc(r->nvertices * sizeof(vec3));
	if (!n) return(NULL);
	
	for (i = 0; i < r->nvertices; i++) {
		vec3_set_all(&n[i], 0.0);
	}
	
	for (t = 0; t < r->nfaces; t++) {
		v0n = r->faces[t].v[0];
		v1n = r->faces[t].v[1];
		v2n = r->faces[t].v[2];
		
		vec3_difference(&v1, &r->vertices[v1n], &r->vertices[v0n]);
		vec3_difference(&v2, &r->vertices[v2n], &r->vertices[v0n]);
		
		vec3_cross_product(&vc, &v1, &v2);
		vec3_normalize(&vc);
		
		vec3_add(&n[v0n], &vc);
		vec3_add(&n[v1n], &vc);
		vec3_add(&n[v2n], &vc);
	}
	
	for (i = 0; i < r->nvertices; i++) {
		vec3_normalize(&n[i]);
		// this is quite suboptimal, change this if we need performance
	}
	return(n);
}



