/*
    This file is part of the lin-kd-tree library.
    Copyright (C) 2009  Laurent Poirrier

    The lin-kd-tree library is free software: you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public License as
    published by the Free Software Foundation, version 3 of the License.

    The lin-kd-tree library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
    General Public License for more details.

    You should have received a copy of the GNU Lesser General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include <math.h>
#include "space.h"

// ***********************************************************************
// 
// ***********************************************************************
void kdtree_set_space(kdtree_space *s,
	KDTREE_SCALAR x0, KDTREE_SCALAR x1, KDTREE_SCALAR y0, KDTREE_SCALAR y1,
	KDTREE_SCALAR z0, KDTREE_SCALAR z1)
{
	KDTREE_SCALAR max = pow(2.0, KDTREE_COORD_BITS);
	
	s->x0 = x0;
	s->y0 = y0;
	s->z0 = z0;
	s->x_width = x1 - x0;
	s->y_width = y1 - y0;
	s->z_width = z1 - z0;
	
	s->x_mult = max / s->x_width;
	s->y_mult = max / s->y_width;
	s->z_mult = max / s->z_width;
}

void kdtree_set_isospace(kdtree_space *s,
	KDTREE_SCALAR x0, KDTREE_SCALAR x1, KDTREE_SCALAR y0, KDTREE_SCALAR y1,
	KDTREE_SCALAR z0, KDTREE_SCALAR z1)
{
	KDTREE_SCALAR cx, cy, cz;
	KDTREE_SCALAR dim;
	
	// keep center 1/2 * ((x0, y0, z0) + (x1, y1; z1))
	dim = x1 - x0;
	if (y1 - y0 > dim) dim = y1 - y0;
	if (z1 - z0 > dim) dim = z1 - z0;
	dim *= 0.5;
	
	cx = 0.5 * (x0 + x1);
	cy = 0.5 * (y0 + y1);
	cz = 0.5 * (z0 + z1);

	kdtree_set_space(s,
		cx - dim, cx + dim, cy - dim, cy + dim, cz - dim, cz + dim);
}

/*
void kdtree_set_isospace(kdtree_space *s,
	KDTREE_SCALAR x0, KDTREE_SCALAR x1, KDTREE_SCALAR y0, KDTREE_SCALAR y1,
	KDTREE_SCALAR z0, KDTREE_SCALAR z1)
{
	// keep lower boundary (x0, y0, z0)
	
	kdtree_set_space(s, x0, x1, y0, y1, z0, z1);
	
	if ((s->x_mult < s->y_mult) && (s->x_mult < s->z_mult)) {
		s->y_mult = s->z_mult = s->x_mult;
	} else if (s->y_mult < s->z_mult) {
		s->x_mult = s->z_mult = s->y_mult;
	} else {
		s->x_mult = s->y_mult = s->z_mult;
	}
}
*/

// ***********************************************************************
// 
// ***********************************************************************
int kdtree_pack_aabb(kdtree_space *s,
	kdtree_iaabb *dst, kdtree_aabb *src)
{
	KDTREE_SCALAR x0, x1, y0, y1, z0, z1;
	
	x0 = src->x0 - s->x0;
	x1 = src->x1 - s->x0;
	if ((x0 < 0.0) || (x1 > s->x_width))
		return(-1);
	
	y0 = src->y0 - s->y0;
	y1 = src->y1 - s->y0;
	if ((y0 < 0.0) || (y1 > s->y_width))
		return(-1);

	z0 = src->z0 - s->z0;
	z1 = src->z1 - s->z0;
	if ((z0 < 0.0) || (z1 > s->z_width))
		return(-1);

	dst->x0 = floor(x0 * s->x_mult);
	dst->x1 = floor(x1 * s->x_mult);
	dst->y0 = floor(y0 * s->y_mult);
	dst->y1 = floor(y1 * s->y_mult);
	dst->z0 = floor(z0 * s->z_mult);
	dst->z1 = floor(z1 * s->z_mult);

	return(0);
}

// ***********************************************************************
// 
// ***********************************************************************
int kdtree_pack_sphere(kdtree_space *s,
	kdtree_iaabb *dst, kdtree_sphere *src)
{
	KDTREE_SCALAR x0, x1, y0, y1, z0, z1;
	
	x0 = src->x - src->r - s->x0;
	x1 = src->x + src->r - s->x0;
	if ((x0 < 0.0) || (x1 > s->x_width))
		return(-1);
	
	y0 = src->y - src->r - s->y0;
	y1 = src->y + src->r - s->y0;
	if ((y0 < 0.0) || (y1 > s->y_width))
		return(-1);

	z0 = src->z - src->r - s->z0;
	z1 = src->z + src->r - s->z0;
	if ((z0 < 0.0) || (z1 > s->z_width))
		return(-1);

	dst->x0 = floor(x0 * s->x_mult);
	dst->x1 =  ceil(x1 * s->x_mult); // shouldn't it be floor here?
	dst->y0 = floor(y0 * s->y_mult);
	dst->y1 =  ceil(y1 * s->y_mult);
	dst->z0 = floor(z0 * s->z_mult);
	dst->z1 =  ceil(z1 * s->z_mult);
	
	return(0);
}

// ***********************************************************************
// Debugging
// ***********************************************************************
void kdtree_retreive_aabb(kdtree_space *s,
	kdtree_iaabb *bb, kdtree_aabb *dst)
{
	dst->x0 = ((KDTREE_SCALAR)bb->x0 / s->x_mult) + s->x0;
	dst->x1 = ((KDTREE_SCALAR)bb->x1 / s->x_mult) + s->x0;
	dst->y0 = ((KDTREE_SCALAR)bb->y0 / s->y_mult) + s->y0;
	dst->y1 = ((KDTREE_SCALAR)bb->y1 / s->y_mult) + s->y0;
	dst->z0 = ((KDTREE_SCALAR)bb->z0 / s->z_mult) + s->z0;
	dst->z1 = ((KDTREE_SCALAR)bb->z1 / s->z_mult) + s->z0;
}

void kdtree_node_aabb(kdtree_space *s, kdtree_node *k, kdtree_aabb *dst)
{
	kdtree_iaabb bb;
	
	kdtree_node_iaabb(k, &bb);
	kdtree_retreive_aabb(s, &bb, dst);
}


